@@ -27,7 +27,7 @@
                                   Just one little problem:
                                   core keeps dumping.''
                                             -- Unknown    */
-#include "mod_ssl.h"
+#include "ssl_private.h"
 
 /*  _________________________________________________________________
 **
@@ -64,10 +64,10 @@
  * - The char* is placed in a brigade and returned
  *
  * Since connection-level input filters in httpd need to be able to
- * handle AP_MODE_GETLINE calls (namely identifying LF-terminated strings), 
+ * handle AP_MODE_GETLINE calls (namely identifying LF-terminated strings),
  * ssl_io_input_getline which will handle this special case.
  *
- * Due to AP_MODE_GETLINE and AP_MODE_SPECULATIVE, we may sometimes have 
+ * Due to AP_MODE_GETLINE and AP_MODE_SPECULATIVE, we may sometimes have
  * 'leftover' decoded data which must be setaside for the next read.  That
  * is currently handled by the char_buffer_{read|write} functions.  So,
  * ssl_io_input_read may be able to fulfill reads without invoking
@@ -81,7 +81,7 @@
  *
  */
 
-/* this custom BIO allows us to hook SSL_write directly into 
+/* this custom BIO allows us to hook SSL_write directly into
  * an apr_bucket_brigade and use transient buckets with the SSL
  * malloc-ed buffer, rather than copying into a mem BIO.
  * also allows us to pass the brigade as data is being written
@@ -102,7 +102,6 @@ typedef struct {
     ap_filter_t        *pInputFilter;
     ap_filter_t        *pOutputFilter;
     int                nobuffer; /* non-zero to prevent buffering */
-    SSLConnRec         *config;
 } ssl_filter_ctx_t;
 
 typedef struct {
@@ -142,7 +141,7 @@ static int bio_filter_out_flush(BIO *bio)
     if (outctx->blen) {
         e = apr_bucket_transient_create(outctx->buffer, outctx->blen,
                                         outctx->bb->bucket_alloc);
-        /* we filled this buffer first so add it to the 
+        /* we filled this buffer first so add it to the
          * head of the brigade
          */
         APR_BRIGADE_INSERT_HEAD(outctx->bb, e);
@@ -183,7 +182,7 @@ static int bio_filter_destroy(BIO *bio)
      */
     return 1;
 }
-	
+
 static int bio_filter_out_read(BIO *bio, char *out, int outl)
 {
     /* this is never called */
@@ -193,13 +192,7 @@ static int bio_filter_out_read(BIO *bio, char *out, int outl)
 static int bio_filter_out_write(BIO *bio, const char *in, int inl)
 {
     bio_filter_out_ctx_t *outctx = (bio_filter_out_ctx_t *)(bio->ptr);
-    
-    /* Abort early if the client has initiated a renegotiation. */
-    if (outctx->filter_ctx->config->reneg_state == RENEG_ABORT) {
-        outctx->rc = APR_ECONNABORTED;
-        return -1;
-    }
-    
+
     /* when handshaking we'll have a small number of bytes.
      * max size SSL will pass us here is about 16k.
      * (16413 bytes to be exact)
@@ -220,7 +213,7 @@ static int bio_filter_out_write(BIO *bio, const char *in, int inl)
     }
     else {
         /* pass along the encrypted data
-         * need to flush since we're using SSL's malloc-ed buffer 
+         * need to flush since we're using SSL's malloc-ed buffer
          * which will be overwritten once we leave here
          */
         apr_bucket *bucket = apr_bucket_transient_create(in, inl,
@@ -343,13 +336,6 @@ typedef struct {
  * this char_buffer api might seem silly, but we don't need to copy
  * any of this data and we need to remember the length.
  */
-
-/* Copy up to INL bytes from the char_buffer BUFFER into IN.  Note
- * that due to the strange way this API is designed/used, the
- * char_buffer object is used to cache a segment of inctx->buffer, and
- * then this function called to copy (part of) that segment to the
- * beginning of inctx->buffer.  So the segments to copy cannot be
- * presumed to be non-overlapping, and memmove must be used. */
 static int char_buffer_read(char_buffer_t *buffer, char *in, int inl)
 {
     if (!buffer->length) {
@@ -358,13 +344,13 @@ static int char_buffer_read(char_buffer_t *buffer, char *in, int inl)
 
     if (buffer->length > inl) {
         /* we have have enough to fill the caller's buffer */
-        memmove(in, buffer->value, inl);
+        memcpy(in, buffer->value, inl);
         buffer->value += inl;
         buffer->length -= inl;
     }
     else {
         /* swallow remainder of the buffer */
-        memmove(in, buffer->value, buffer->length);
+        memcpy(in, buffer->value, buffer->length);
         inl = buffer->length;
         buffer->value = NULL;
         buffer->length = 0;
@@ -389,7 +375,7 @@ static apr_status_t brigade_consume(apr_bucket_brigade *bb,
 {
     apr_size_t actual = 0;
     apr_status_t status = APR_SUCCESS;
- 
+
     while (!APR_BRIGADE_EMPTY(bb)) {
         apr_bucket *b = APR_BRIGADE_FIRST(bb);
         const char *str;
@@ -412,11 +398,11 @@ static apr_status_t brigade_consume(apr_bucket_brigade *bb,
          * this sort of case (caller provided buffer) it
          * would be much more trivial for apr_bucket_consume
          * to do all the work that follows, based on the
-         * particular characteristics of the bucket we are 
+         * particular characteristics of the bucket we are
          * consuming here.
          */
         status = apr_bucket_read(b, &str, &str_len, block);
-        
+
         if (status != APR_SUCCESS) {
             if (APR_STATUS_IS_EOF(status)) {
                 /* This stream bucket was consumed */
@@ -479,19 +465,13 @@ static int bio_filter_in_read(BIO *bio, char *in, int inlen)
     if (!in)
         return 0;
 
-    /* Abort early if the client has initiated a renegotiation. */
-    if (inctx->filter_ctx->config->reneg_state == RENEG_ABORT) {
-        inctx->rc = APR_ECONNABORTED;
-        return -1;
-    }
-
     /* XXX: flush here only required for SSLv2;
      * OpenSSL calls BIO_flush() at the appropriate times for
      * the other protocols.
      */
     if ((SSL_version(inctx->ssl) == SSL2_VERSION) || sslconn->is_proxy) {
         if (bio_filter_out_flush(inctx->bio_out) < 0) {
-            bio_filter_out_ctx_t *outctx = 
+            bio_filter_out_ctx_t *outctx =
                    (bio_filter_out_ctx_t *)(inctx->bio_out->ptr);
             inctx->rc = outctx->rc;
             return -1;
@@ -508,15 +488,17 @@ static int bio_filter_in_read(BIO *bio, char *in, int inlen)
     if (APR_BRIGADE_EMPTY(inctx->bb)) {
 
         inctx->rc = ap_get_brigade(inctx->f->next, inctx->bb,
-                                   AP_MODE_READBYTES, block, 
+                                   AP_MODE_READBYTES, block,
                                    inl);
 
-        /* Not a problem, there was simply no data ready yet.
-         */
+        /* If the read returns EAGAIN or success with an empty
+         * brigade, return an error after setting the retry flag;
+         * SSL_read() will then return -1, and SSL_get_error() will
+         * indicate SSL_ERROR_WANT_READ. */
         if (APR_STATUS_IS_EAGAIN(inctx->rc) || APR_STATUS_IS_EINTR(inctx->rc)
                || (inctx->rc == APR_SUCCESS && APR_BRIGADE_EMPTY(inctx->bb))) {
             BIO_set_retry_read(bio);
-            return 0;
+            return -1;
         }
 
         if (inctx->rc != APR_SUCCESS) {
@@ -533,12 +515,12 @@ static int bio_filter_in_read(BIO *bio, char *in, int inlen)
         return (int)inl;
     }
 
-    if (APR_STATUS_IS_EAGAIN(inctx->rc) 
+    if (APR_STATUS_IS_EAGAIN(inctx->rc)
             || APR_STATUS_IS_EINTR(inctx->rc)) {
         BIO_set_retry_read(bio);
         return (int)inl;
     }
-        
+
     /* Unexpected errors and APR_EOF clean out the brigade.
      * Subsequent calls will return APR_EOF.
      */
@@ -643,7 +625,7 @@ static apr_status_t ssl_io_input_read(bio_filter_in_ctx_t *inctx,
              */
             if (APR_STATUS_IS_EAGAIN(inctx->rc)
                     || APR_STATUS_IS_EINTR(inctx->rc)) {
-                /* Already read something, return APR_SUCCESS instead. 
+                /* Already read something, return APR_SUCCESS instead.
                  * On win32 in particular, but perhaps on other kernels,
                  * a blocking call isn't 'always' blocking.
                  */
@@ -703,16 +685,16 @@ static apr_status_t ssl_io_input_read(bio_filter_in_ctx_t *inctx,
                     continue;  /* Blocking and nothing yet?  Try again. */
                 }
                 else {
-                    ap_log_error(APLOG_MARK, APLOG_INFO, inctx->rc, c->base_server,
-                                "SSL input filter read failed.");
+                    ap_log_cerror(APLOG_MARK, APLOG_INFO, inctx->rc, c,
+                                  "SSL input filter read failed.");
                 }
             }
             else /* if (ssl_err == SSL_ERROR_SSL) */ {
                 /*
                  * Log SSL errors and any unexpected conditions.
                  */
-                ap_log_error(APLOG_MARK, APLOG_INFO, inctx->rc, c->base_server,
-                            "SSL library error %d reading data", ssl_err);
+                ap_log_cerror(APLOG_MARK, APLOG_INFO, inctx->rc, c,
+                              "SSL library error %d reading data", ssl_err);
                 ssl_log_ssl_error(APLOG_MARK, APLOG_INFO, c->base_server);
 
             }
@@ -743,7 +725,7 @@ static apr_status_t ssl_io_input_getline(bio_filter_in_ctx_t *inctx,
 
     while (tmplen > 0) {
         status = ssl_io_input_read(inctx, buf + offset, &tmplen);
-        
+
         if (status != APR_SUCCESS) {
             return status;
         }
@@ -808,15 +790,15 @@ static apr_status_t ssl_filter_write(ap_filter_t *f,
             outctx->rc = APR_EAGAIN;
         }
         else if (ssl_err == SSL_ERROR_SYSCALL) {
-            ap_log_error(APLOG_MARK, APLOG_INFO, outctx->rc, c->base_server,
-                        "SSL output filter write failed.");
+            ap_log_cerror(APLOG_MARK, APLOG_INFO, outctx->rc, c,
+                          "SSL output filter write failed.");
         }
         else /* if (ssl_err == SSL_ERROR_SSL) */ {
             /*
              * Log SSL errors
              */
-            ap_log_error(APLOG_MARK, APLOG_INFO, outctx->rc, c->base_server,
-                         "SSL library error %d writing data", ssl_err);
+            ap_log_cerror(APLOG_MARK, APLOG_INFO, outctx->rc, c,
+                          "SSL library error %d writing data", ssl_err);
             ssl_log_ssl_error(APLOG_MARK, APLOG_INFO, c->base_server);
         }
         if (outctx->rc == APR_SUCCESS) {
@@ -832,9 +814,10 @@ static apr_status_t ssl_filter_write(ap_filter_t *f,
             reason = "likely due to failed renegotiation";
         }
 
-        ap_log_error(APLOG_MARK, APLOG_INFO, outctx->rc, c->base_server,
-                     "failed to write %d of %d bytes (%s)",
-                     len - (apr_size_t)res, len, reason);
+        ap_log_cerror(APLOG_MARK, APLOG_INFO, outctx->rc, c,
+                      "failed to write %" APR_SSIZE_T_FMT
+                      " of %" APR_SIZE_T_FMT " bytes (%s)",
+                      len - (apr_size_t)res, len, reason);
 
         outctx->rc = APR_EGENERAL;
     }
@@ -875,8 +858,7 @@ static apr_status_t ssl_io_filter_error(ap_filter_t *f,
     switch (status) {
       case HTTP_BAD_REQUEST:
             /* log the situation */
-            ap_log_error(APLOG_MARK, APLOG_INFO, 0,
-                         f->c->base_server,
+            ap_log_cerror(APLOG_MARK, APLOG_INFO, 0, f->c,
                          "SSL handshake failed: HTTP spoken on HTTPS port; "
                          "trying to send HTML error page");
             ssl_log_ssl_error(APLOG_MARK, APLOG_INFO, f->c->base_server);
@@ -905,7 +887,7 @@ static const char ssl_io_buffer[] = "SSL/TLS Buffer";
 /*
  *  Close the SSL part of the socket connection
  *  (called immediately _before_ the socket is closed)
- *  or called with 
+ *  or called with
  */
 static apr_status_t ssl_filter_io_shutdown(ssl_filter_ctx_t *filter_ctx,
                                            conn_rec *c,
@@ -990,12 +972,10 @@ static apr_status_t ssl_filter_io_shutdown(ssl_filter_ctx_t *filter_ctx,
 
     /* and finally log the fact that we've closed the connection */
     if (c->base_server->loglevel >= APLOG_INFO) {
-        ap_log_error(APLOG_MARK, APLOG_INFO, 0, c->base_server,
-                     "Connection to child %ld closed with %s shutdown"
-                     "(server %s, client %s)",
-                     c->id, type,
-                     ssl_util_vhostid(c->pool, c->base_server),
-                     c->remote_ip ? c->remote_ip : "unknown");
+        ap_log_cerror(APLOG_MARK, APLOG_INFO, 0, c,
+                      "Connection closed to child %ld with %s shutdown "
+                      "(server %s)",
+                      c->id, type, ssl_util_vhostid(c->pool, c->base_server));
     }
 
     /* deallocate the SSL connection */
@@ -1023,13 +1003,10 @@ static apr_status_t ssl_io_filter_cleanup(void *data)
         conn_rec *c = (conn_rec *)SSL_get_app_data(filter_ctx->pssl);
         SSLConnRec *sslconn = myConnConfig(c);
 
-        ap_log_error(APLOG_MARK, APLOG_DEBUG, 0, NULL,
-                     "SSL connection destroyed without being closed");
-
         SSL_free(filter_ctx->pssl);
         sslconn->ssl = filter_ctx->pssl = NULL;
     }
-  
+
     return APR_SUCCESS;
 }
 
@@ -1056,11 +1033,12 @@ static int ssl_io_filter_connect(ssl_filter_ctx_t *filter_ctx)
 
     if (sslconn->is_proxy) {
         if ((n = SSL_connect(filter_ctx->pssl)) <= 0) {
-            ap_log_error(APLOG_MARK, APLOG_INFO, 0,
-                         c->base_server,
-                         "SSL Proxy connect failed");
+            ap_log_cerror(APLOG_MARK, APLOG_INFO, 0, c,
+                          "SSL Proxy connect failed");
             ssl_log_ssl_error(APLOG_MARK, APLOG_INFO, c->base_server);
-            return ssl_filter_io_shutdown(filter_ctx, c, 1);
+            /* ensure that the SSL structures etc are freed, etc: */
+            ssl_filter_io_shutdown(filter_ctx, c, 1);
+            return HTTP_BAD_GATEWAY;
         }
 
         return APR_SUCCESS;
@@ -1073,20 +1051,19 @@ static int ssl_io_filter_connect(ssl_filter_ctx_t *filter_ctx)
                                        (filter_ctx->pbioWrite->ptr);
         apr_status_t rc = inctx->rc ? inctx->rc : outctx->rc ;
         ssl_err = SSL_get_error(filter_ctx->pssl, n);
-        
+
         if (ssl_err == SSL_ERROR_ZERO_RETURN) {
             /*
              * The case where the connection was closed before any data
              * was transferred. That's not a real error and can occur
              * sporadically with some clients.
              */
-            ap_log_error(APLOG_MARK, APLOG_INFO, rc,
-                         c->base_server,
+            ap_log_cerror(APLOG_MARK, APLOG_INFO, rc, c,
                          "SSL handshake stopped: connection was closed");
         }
         else if (ssl_err == SSL_ERROR_WANT_READ) {
             /*
-             * This is in addition to what was present earlier. It is 
+             * This is in addition to what was present earlier. It is
              * borrowed from openssl_state_machine.c [mod_tls].
              * TBD.
              */
@@ -1104,19 +1081,18 @@ static int ssl_io_filter_connect(ssl_filter_ctx_t *filter_ctx)
             return HTTP_BAD_REQUEST;
         }
         else if (ssl_err == SSL_ERROR_SYSCALL) {
-            ap_log_error(APLOG_MARK, APLOG_INFO, rc, c->base_server,
-                         "SSL handshake interrupted by system "
-                         "[Hint: Stop button pressed in browser?!]");
+            ap_log_cerror(APLOG_MARK, APLOG_INFO, rc, c,
+                          "SSL handshake interrupted by system "
+                          "[Hint: Stop button pressed in browser?!]");
         }
         else /* if (ssl_err == SSL_ERROR_SSL) */ {
             /*
              * Log SSL errors and any unexpected conditions.
              */
-            ap_log_error(APLOG_MARK, APLOG_INFO, rc, c->base_server,
-                         "SSL library error %d in handshake "
-                         "(server %s, client %s)", ssl_err,
-                         ssl_util_vhostid(c->pool, c->base_server),
-                         c->remote_ip ? c->remote_ip : "unknown");
+            ap_log_cerror(APLOG_MARK, APLOG_INFO, rc, c,
+                          "SSL library error %d in handshake "
+                          "(server %s)", ssl_err,
+                          ssl_util_vhostid(c->pool, c->base_server));
             ssl_log_ssl_error(APLOG_MARK, APLOG_INFO, c->base_server);
 
         }
@@ -1141,17 +1117,16 @@ static int ssl_io_filter_connect(ssl_filter_ctx_t *filter_ctx)
             /* leaving this log message as an error for the moment,
              * according to the mod_ssl docs:
              * "level optional_no_ca is actually against the idea
-             *  of authentication (but can be used to establish 
+             *  of authentication (but can be used to establish
              * SSL test pages, etc.)"
              * optional_no_ca doesn't appear to work as advertised
              * in 1.x
              */
-            ap_log_error(APLOG_MARK, APLOG_INFO, 0,
-                         c->base_server,
-                         "SSL client authentication failed, "
-                         "accepting certificate based on "
-                         "\"SSLVerifyClient optional_no_ca\" "
-                         "configuration");
+            ap_log_cerror(APLOG_MARK, APLOG_INFO, 0, c,
+                          "SSL client authentication failed, "
+                          "accepting certificate based on "
+                          "\"SSLVerifyClient optional_no_ca\" "
+                          "configuration");
             ssl_log_ssl_error(APLOG_MARK, APLOG_INFO, c->base_server);
         }
         else {
@@ -1159,8 +1134,7 @@ static int ssl_io_filter_connect(ssl_filter_ctx_t *filter_ctx)
                 sslconn->verify_error :
                 X509_verify_cert_error_string(verify_result);
 
-            ap_log_error(APLOG_MARK, APLOG_INFO, 0,
-                         c->base_server,
+            ap_log_cerror(APLOG_MARK, APLOG_INFO, 0, c,
                          "SSL client authentication failed: %s",
                          error ? error : "unknown");
             ssl_log_ssl_error(APLOG_MARK, APLOG_INFO, c->base_server);
@@ -1187,8 +1161,8 @@ static int ssl_io_filter_connect(ssl_filter_ctx_t *filter_ctx)
     if ((sc->server->auth.verify_mode == SSL_CVERIFY_REQUIRE) &&
         !sslconn->client_cert)
     {
-        ap_log_error(APLOG_MARK, APLOG_INFO, 0, c->base_server,
-                     "No acceptable peer certificate available");
+        ap_log_cerror(APLOG_MARK, APLOG_INFO, 0, c,
+                      "No acceptable peer certificate available");
 
         return ssl_filter_io_shutdown(filter_ctx, c, 1);
     }
@@ -1196,6 +1170,85 @@ static int ssl_io_filter_connect(ssl_filter_ctx_t *filter_ctx)
     return APR_SUCCESS;
 }
 
+#define SWITCH_STATUS_LINE "HTTP/1.1 101 Switching Protocols"
+#define UPGRADE_HEADER "Upgrade: TLS/1.0, HTTP/1.1"
+#define CONNECTION_HEADER "Connection: Upgrade"
+
+static apr_status_t ssl_io_filter_Upgrade(ap_filter_t *f,
+                                          apr_bucket_brigade *bb)
+{
+    const char *upgrade;
+    apr_bucket_brigade *upgradebb;
+    request_rec *r = f->r;
+    SSLConnRec *sslconn;
+    apr_status_t rv;
+    apr_bucket *b;
+    SSL *ssl;
+
+    /* Just remove the filter, if it doesn't work the first time, it won't
+     * work at all for this request.
+     */
+    ap_remove_output_filter(f);
+
+    /* No need to ensure that this is a server with optional SSL, the filter
+     * is only inserted if that is true.
+     */
+
+    upgrade = apr_table_get(r->headers_in, "Upgrade");
+    if (upgrade == NULL
+        || strcmp(ap_getword(r->pool, &upgrade, ','), "TLS/1.0")) {
+        /* "Upgrade: TLS/1.0, ..." header not found, don't do Upgrade */
+        return ap_pass_brigade(f->next, bb);
+    }
+
+    apr_table_unset(r->headers_out, "Upgrade");
+
+    /* Send the interim 101 response. */
+    upgradebb = apr_brigade_create(r->pool, f->c->bucket_alloc);
+
+    ap_fputstrs(f->next, upgradebb, SWITCH_STATUS_LINE, CRLF,
+                UPGRADE_HEADER, CRLF, CONNECTION_HEADER, CRLF, CRLF, NULL);
+
+    b = apr_bucket_flush_create(f->c->bucket_alloc);
+    APR_BRIGADE_INSERT_TAIL(upgradebb, b);
+
+    rv = ap_pass_brigade(f->next, upgradebb);
+    if (rv) {
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
+                      "could not send interim 101 Upgrade response");
+        return AP_FILTER_ERROR;
+    }
+
+    ssl_init_ssl_connection(f->c);
+
+    ap_log_rerror(APLOG_MARK, APLOG_INFO, 0, r,
+                  "Awaiting re-negotiation handshake");
+
+    sslconn = myConnConfig(f->c);
+    ssl = sslconn->ssl;
+
+    /* XXX: Should replace SSL_set_state with SSL_renegotiate(ssl);
+     * However, this causes failures in perl-framework currently,
+     * perhaps pre-test if we have already negotiated?
+     */
+    SSL_set_accept_state(ssl);
+    SSL_do_handshake(ssl);
+
+    if (SSL_get_state(ssl) != SSL_ST_OK) {
+        ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+                      "TLS Upgrade handshake failed: "
+                      "Not accepted by client!?");
+
+        return AP_FILTER_ERROR;
+    }
+
+    /* Now that we have initialized the ssl connection which added the ssl_io_filter,
+       pass the brigade off to the connection based output filters so that the
+       request can complete encrypted */
+    return ap_pass_brigade(f->c->output_filters, bb);
+
+}
+
 static apr_status_t ssl_io_filter_input(ap_filter_t *f,
                                         apr_bucket_brigade *bb,
                                         ap_input_mode_t mode,
@@ -1223,7 +1276,7 @@ static apr_status_t ssl_io_filter_input(ap_filter_t *f,
     }
 
     /* XXX: we don't currently support anything other than these modes. */
-    if (mode != AP_MODE_READBYTES && mode != AP_MODE_GETLINE && 
+    if (mode != AP_MODE_READBYTES && mode != AP_MODE_GETLINE &&
         mode != AP_MODE_SPECULATIVE && mode != AP_MODE_INIT) {
         return APR_ENOTIMPL;
     }
@@ -1247,9 +1300,9 @@ static apr_status_t ssl_io_filter_input(ap_filter_t *f,
         return APR_SUCCESS;
     }
 
-    if (inctx->mode == AP_MODE_READBYTES || 
+    if (inctx->mode == AP_MODE_READBYTES ||
         inctx->mode == AP_MODE_SPECULATIVE) {
-        /* Protected from truncation, readbytes < MAX_SIZE_T 
+        /* Protected from truncation, readbytes < MAX_SIZE_T
          * FIXME: No, it's *not* protected.  -- jre */
         if (readbytes < len) {
             len = (apr_size_t)readbytes;
@@ -1261,17 +1314,9 @@ static apr_status_t ssl_io_filter_input(ap_filter_t *f,
     }
     else {
         /* We have no idea what you are talking about, so return an error. */
-        status = APR_ENOTIMPL;
+        return APR_ENOTIMPL;
     }
 
-    /* It is possible for mod_ssl's BIO to be used outside of the
-     * direct control of mod_ssl's input or output filter -- notably,
-     * when mod_ssl initiates a renegotiation.  Switching the BIO mode
-     * back to "blocking" here ensures such operations don't fail with
-     * SSL_ERROR_WANT_READ. */
-    inctx->block = APR_BLOCK_READ;
-
-    /* Handle custom errors. */
     if (status != APR_SUCCESS) {
         return ssl_io_filter_error(f, bb, status);
     }
@@ -1322,8 +1367,8 @@ static apr_status_t ssl_io_filter_output(ap_filter_t *f,
     while (!APR_BRIGADE_EMPTY(bb)) {
         apr_bucket *bucket = APR_BRIGADE_FIRST(bb);
 
-        /* If it is a flush or EOS, we need to pass this down. 
-         * These types do not require translation by OpenSSL.  
+        /* If it is a flush or EOS, we need to pass this down.
+         * These types do not require translation by OpenSSL.
          */
         if (APR_BUCKET_IS_EOS(bucket) || APR_BUCKET_IS_FLUSH(bucket)) {
             if (bio_filter_out_flush(filter_ctx->pbioWrite) < 0) {
@@ -1359,8 +1404,8 @@ static apr_status_t ssl_io_filter_output(ap_filter_t *f,
             filter_ctx->nobuffer = 1;
             status = ssl_filter_io_shutdown(filter_ctx, f->c, 0);
             if (status != APR_SUCCESS) {
-                ap_log_error(APLOG_MARK, APLOG_INFO, status, NULL,
-                             "SSL filter error shutting down I/O");
+                ap_log_cerror(APLOG_MARK, APLOG_INFO, status, f->c,
+                              "SSL filter error shutting down I/O");
             }
             if ((status = ap_pass_brigade(f->next, bb)) != APR_SUCCESS) {
                 return status;
@@ -1371,7 +1416,7 @@ static apr_status_t ssl_io_filter_output(ap_filter_t *f,
             /* filter output */
             const char *data;
             apr_size_t len;
-            
+
             status = apr_bucket_read(bucket, &data, &len, rblock);
 
             if (APR_STATUS_IS_EAGAIN(status)) {
@@ -1419,7 +1464,7 @@ int ssl_io_buffer_fill(request_rec *r)
     apr_bucket_brigade *tempb;
     apr_off_t total = 0; /* total length buffered */
     int eos = 0; /* non-zero once EOS is seen */
-    
+
     /* Create the context which will be passed to the input filter;
      * containing a setaside pool and a brigade which constrain the
      * lifetime of the buffered data. */
@@ -1430,7 +1475,7 @@ int ssl_io_buffer_fill(request_rec *r)
     /* ... and a temporary brigade. */
     tempb = apr_brigade_create(r->pool, c->bucket_alloc);
 
-    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, "filling buffer");
+    ap_log_cerror(APLOG_MARK, APLOG_DEBUG, 0, c, "filling buffer");
 
     do {
         apr_status_t rv;
@@ -1448,10 +1493,10 @@ int ssl_io_buffer_fill(request_rec *r)
                           "could not read request body for SSL buffer");
             return HTTP_INTERNAL_SERVER_ERROR;
         }
-        
+
         /* Iterate through the returned brigade: setaside each bucket
          * into the context's pool and move it into the brigade. */
-        for (e = APR_BRIGADE_FIRST(tempb); 
+        for (e = APR_BRIGADE_FIRST(tempb);
              e != APR_BRIGADE_SENTINEL(tempb) && !eos; e = next) {
             const char *data;
             apr_size_t len;
@@ -1469,19 +1514,19 @@ int ssl_io_buffer_fill(request_rec *r)
                 }
                 total += len;
             }
-                
+
             rv = apr_bucket_setaside(e, ctx->pool);
             if (rv != APR_SUCCESS) {
                 ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
                               "could not setaside bucket for SSL buffer");
                 return HTTP_INTERNAL_SERVER_ERROR;
             }
-            
+
             APR_BUCKET_REMOVE(e);
             APR_BRIGADE_INSERT_TAIL(ctx->bb, e);
         }
 
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r, 
+        ap_log_cerror(APLOG_MARK, APLOG_DEBUG, 0, c,
                       "total of %" APR_OFF_T_FMT " bytes in buffer, eos=%d",
                       total, eos);
 
@@ -1513,7 +1558,7 @@ static apr_status_t ssl_io_filter_buffer(ap_filter_t *f,
     struct modssl_buffer_ctx *ctx = f->ctx;
     apr_status_t rv;
 
-    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
+    ap_log_cerror(APLOG_MARK, APLOG_DEBUG, 0, f->c,
                   "read from buffered SSL brigade, mode %d, "
                   "%" APR_OFF_T_FMT " bytes",
                   mode, bytes);
@@ -1528,7 +1573,7 @@ static apr_status_t ssl_io_filter_buffer(ap_filter_t *f,
         /* Partition the buffered brigade. */
         rv = apr_brigade_partition(ctx->bb, bytes, &e);
         if (rv && rv != APR_INCOMPLETE) {
-            ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, f->r,
+            ap_log_cerror(APLOG_MARK, APLOG_ERR, rv, f->c,
                           "could not partition buffered SSL brigade");
             ap_remove_input_filter(f);
             return rv;
@@ -1542,7 +1587,7 @@ static apr_status_t ssl_io_filter_buffer(ap_filter_t *f,
             apr_bucket *d = APR_BRIGADE_FIRST(ctx->bb);
 
             e = APR_BUCKET_PREV(e);
-            
+
             /* Unsplice the partitioned segment and move it into the
              * passed-in brigade; no convenient way to do this with
              * the APR_BRIGADE_* macros. */
@@ -1558,7 +1603,7 @@ static apr_status_t ssl_io_filter_buffer(ap_filter_t *f,
         rv = apr_brigade_split_line(bb, ctx->bb, mode, bytes);
 
         if (rv) {
-            ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, f->r,
+            ap_log_cerror(APLOG_MARK, APLOG_ERR, rv, f->c,
                           "could not split line from buffered SSL brigade");
             ap_remove_input_filter(f);
             return rv;
@@ -1567,7 +1612,7 @@ static apr_status_t ssl_io_filter_buffer(ap_filter_t *f,
 
     if (APR_BRIGADE_EMPTY(ctx->bb)) {
         apr_bucket *e = APR_BRIGADE_LAST(bb);
-        
+
         /* Ensure that the brigade is terminated by an EOS if the
          * buffered request body has been entirely consumed. */
         if (e == APR_BRIGADE_SENTINEL(bb) || !APR_BUCKET_IS_EOS(e)) {
@@ -1575,7 +1620,7 @@ static apr_status_t ssl_io_filter_buffer(ap_filter_t *f,
             APR_BRIGADE_INSERT_TAIL(bb, e);
         }
 
-        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, f->r,
+        ap_log_cerror(APLOG_MARK, APLOG_DEBUG, 0, f->c,
                       "buffered SSL brigade now exhausted; removing filter");
         ap_remove_input_filter(f);
     }
@@ -1613,8 +1658,6 @@ void ssl_io_filter_init(conn_rec *c, SSL *ssl)
 
     filter_ctx = apr_palloc(c->pool, sizeof(ssl_filter_ctx_t));
 
-    filter_ctx->config          = myConnConfig(c);
-
     filter_ctx->nobuffer        = 0;
     filter_ctx->pOutputFilter   = ap_add_output_filter(ssl_io_filter,
                                                    filter_ctx, NULL, c);
@@ -1640,9 +1683,14 @@ void ssl_io_filter_init(conn_rec *c, SSL *ssl)
 
 void ssl_io_filter_register(apr_pool_t *p)
 {
+    /* This filter MUST be after the HTTP_HEADER filter, but it also must be
+     * a resource-level filter so it has the request_rec.
+     */
+    ap_register_output_filter ("UPGRADE_FILTER", ssl_io_filter_Upgrade, NULL, AP_FTYPE_PROTOCOL + 5);
+
     ap_register_input_filter  (ssl_io_filter, ssl_io_filter_input,  NULL, AP_FTYPE_CONNECTION + 5);
     ap_register_output_filter (ssl_io_filter, ssl_io_filter_output, NULL, AP_FTYPE_CONNECTION + 5);
-    
+
     ap_register_input_filter  (ssl_io_buffer, ssl_io_filter_buffer, NULL, AP_FTYPE_PROTOCOL - 1);
 
     return;
