@@ -14,14 +14,22 @@
  * limitations under the License.
  */
 
-#ifndef APACHE_MPM_H
-#define APACHE_MPM_H
-
-/* mpm.h is the place to make declarations that are MPM specific but that must be 
+/**
+ * @file  winnt/mpm.h
+ * @brief MPM for Windows NT
+ * 
+ * this is the place to make declarations that are MPM specific but that must be 
  * shared with non-mpm specific code in the server.  Hummm, perhaps we can
  * move most of this stuff to mpm_common.h?
+ *
+ * @defgroup APACHE_MPM_WINNT WinNT MPM
+ * @ingroup  APACHE_OS_WIN32 APACHE_MPM
+ * @{
  */
 
+#ifndef APACHE_MPM_H
+#define APACHE_MPM_H
+
 #include "scoreboard.h"
 
 #define MPM_NAME "WinNT"
@@ -31,9 +39,11 @@
 #define AP_MPM_WANT_SET_COREDUMPDIR
 #define AP_MPM_WANT_SET_SCOREBOARD
 #define AP_MPM_WANT_SET_MAX_MEM_FREE
+#define AP_MPM_WANT_SET_STACKSIZE
 
 extern int ap_threads_per_child;
 extern int ap_thread_limit;
 extern server_rec *ap_server_conf;
 
 #endif /* APACHE_MPM_H */
+/** @} */
