@@ -20,6 +20,35 @@
 #include "apr_lib.h"
 #include "apr_strings.h"
 
+/* The RMM region is made up of two doubly-linked-list of blocks; the
+ * list of used blocks, and the list of free blocks (either list may
+ * be empty).  The base pointer, rmm->base, points at the beginning of
+ * the shmem region in use.  Each block is addressable by an
+ * apr_rmm_off_t value, which represents the offset from the base
+ * pointer.  The term "address" is used here to mean such a value; an
+ * "offset from rmm->base".
+ *
+ * The RMM region contains exactly one "rmm_hdr_block_t" structure,
+ * the "header block", which is always stored at the base pointer.
+ * The firstused field in this structure is the address of the first
+ * block in the "used blocks" list; the firstfree field is the address
+ * of the first block in the "free blocks" list.
+ *
+ * Each block is prefixed by an "rmm_block_t" structure, followed by
+ * the caller-usable region represented by the block.  The next and
+ * prev fields of the structure are zero if the block is at the end or
+ * beginning of the linked-list respectively, or otherwise hold the
+ * address of the next and previous blocks in the list.  ("address 0",
+ * i.e. rmm->base is *not* a valid address for a block, since the
+ * header block is always stored at that address).
+ *
+ * At creation, the RMM region is initialized to hold a single block
+ * on the free list representing the entire available shm segment
+ * (minus header block); subsequent allocation and deallocation of
+ * blocks involves splitting blocks and coalescing adjacent blocks,
+ * and switching them between the free and used lists as
+ * appropriate. */
+
 typedef struct rmm_block_t {
     apr_size_t size;
     apr_rmm_off_t prev;
@@ -277,17 +306,13 @@ APU_DECLARE(apr_status_t) apr_rmm_detach(apr_rmm_t *rmm)
 
 APU_DECLARE(apr_rmm_off_t) apr_rmm_malloc(apr_rmm_t *rmm, apr_size_t reqsize)
 {
-    apr_size_t size;
     apr_rmm_off_t this;
     
-    size = APR_ALIGN_DEFAULT(reqsize) + RMM_BLOCK_SIZE;
-    if (size < reqsize) {
-        return 0;
-    }
+    reqsize = APR_ALIGN_DEFAULT(reqsize) + RMM_BLOCK_SIZE;
 
     APR_ANYLOCK_LOCK(&rmm->lock);
 
-    this = find_block_of_size(rmm, size);
+    this = find_block_of_size(rmm, reqsize);
 
     if (this) {
         move_block(rmm, this, 0);
@@ -300,22 +325,18 @@ APU_DECLARE(apr_rmm_off_t) apr_rmm_malloc(apr_rmm_t *rmm, apr_size_t reqsize)
 
 APU_DECLARE(apr_rmm_off_t) apr_rmm_calloc(apr_rmm_t *rmm, apr_size_t reqsize)
 {
-    apr_size_t size;
     apr_rmm_off_t this;
         
-    size = APR_ALIGN_DEFAULT(reqsize) + RMM_BLOCK_SIZE;
-    if (size < reqsize) {
-        return 0;
-    }
+    reqsize = APR_ALIGN_DEFAULT(reqsize) + RMM_BLOCK_SIZE;
 
     APR_ANYLOCK_LOCK(&rmm->lock);
 
-    this = find_block_of_size(rmm, size);
+    this = find_block_of_size(rmm, reqsize);
 
     if (this) {
         move_block(rmm, this, 0);
         this += RMM_BLOCK_SIZE;
-        memset((char*)rmm->base + this, 0, size - RMM_BLOCK_SIZE);
+        memset((char*)rmm->base + this, 0, reqsize - RMM_BLOCK_SIZE);
     }
 
     APR_ANYLOCK_UNLOCK(&rmm->lock);
@@ -328,19 +349,16 @@ APU_DECLARE(apr_rmm_off_t) apr_rmm_realloc(apr_rmm_t *rmm, void *entity,
     apr_rmm_off_t this;
     apr_rmm_off_t old;
     struct rmm_block_t *blk;
-    apr_size_t size, oldsize;
+    apr_size_t oldsize;
 
     if (!entity) {
         return apr_rmm_malloc(rmm, reqsize);
     }
 
-    size = APR_ALIGN_DEFAULT(reqsize);
-    if (size < reqsize) {
-        return 0;
-    }
+    reqsize = APR_ALIGN_DEFAULT(reqsize);
     old = apr_rmm_offset_get(rmm, entity);
 
-    if ((this = apr_rmm_malloc(rmm, size)) == 0) {
+    if ((this = apr_rmm_malloc(rmm, reqsize)) == 0) {
         return 0;
     }
 
@@ -348,7 +366,7 @@ APU_DECLARE(apr_rmm_off_t) apr_rmm_realloc(apr_rmm_t *rmm, void *entity,
     oldsize = blk->size;
 
     memcpy(apr_rmm_addr_get(rmm, this),
-           apr_rmm_addr_get(rmm, old), oldsize < size ? oldsize : size);
+           apr_rmm_addr_get(rmm, old), oldsize < reqsize ? oldsize : reqsize);
     apr_rmm_free(rmm, old);
 
     return this;
