@@ -1,10 +1,9 @@
-#include <stdlib.h>
 #include <netware.h>
 #include <screen.h>
+#include "testutil.h"
 
 void _NonAppStop( void )
 {
-  if (getenv("_IN_NETWARE_BASH_") == NULL)
     pressanykey();
 }
 
